import signal
from getpass import getpass

from .core import Miner, ItemMiner, SearchMiner
from .config import write_config_file


def search(query=None, params=None, callback=None, mine_ids=None, info_only=None,
           **kwargs):
    """Mine Archive.org search results.

    :param query: (optional) The Archive.org search query to yield
                  results for. Refer to https://archive.org/advancedsearch.php#raw
                  for help formatting your query. If no query is given,
                  all indexed items will be mined!
    :type query: str

    :param params: (optional) The URL parameters to send with each
                   request sent to the Archive.org Advancedsearch Api.
    :type params: dict

    :param callback: (optional) A callback function to be called on each
                     :py:class:`aiohttp.client.ClientResponse`.

    :param mine_ids: (optional) By default, ``search`` mines through
                     search results. To mine through the item metadata
                     for each item returned by your query instead, set
                     ``mine_ids`` to ``True``.
    :type mine_ids: bool

    :param info_only: (optional) Set to ``True`` to return information
                      about your query rather than mining any metadata
                      or search results.
    :type info_only: bool

    :param \*\*kwargs: (optional) Arguments that ``get_miner`` takes.
    """
    query = '(*:*)' if not query else query
    params = params if params else {}
    mine_ids = True if mine_ids else False
    info_only = True if info_only else False
    miner = SearchMiner(**kwargs)

    if info_only:
        params = miner.get_search_params(query, params)
        r = miner.get_search_info(params)
        search_info = r.get('responseHeader')
        search_info['numFound'] = r.get('response', {}).get('numFound', 0)
        return search_info

    try:
        miner.loop.add_signal_handler(signal.SIGINT, miner.close)
        miner.loop.run_until_complete(
                miner.search(query, params=params, callback=callback, mine_ids=mine_ids))
    except RuntimeError:
        pass


def mine_urls(urls, params=None, callback=None, **kwargs):
    """Concurrently retrieve URLs.

    :param urls: A set of URLs to concurrently retrieve.
    :type urls: iterable

    :param params: (optional) The URL parameters to send with each
                   request.
    :type params: dict

    :param callback: (optional) A callback function to be called on each
                     :py:class:`aiohttp.client.ClientResponse`.

    :param \*\*kwargs: (optional) Arguments that ``get_miner`` takes.
    """
    miner = Miner(**kwargs)
    try:
        miner.loop.add_signal_handler(signal.SIGINT, miner.close)
        miner.loop.run_until_complete(miner.mine_urls(urls, params, callback))
    except RuntimeError:
        pass


def mine_items(identifiers, params=None, callback=None, **kwargs):
    """Concurrently retrieve metadata from Archive.org items.

    :param identifiers: A set of Archive.org item identifiers to mine.
    :type identifiers: iterable

    :param params: (optional) The URL parameters to send with each
                   request.
    :type params: dict

    :param callback: (optional) A callback function to be called on each
                     :py:class:`aiohttp.client.ClientResponse`.

    :param \*\*kwargs: (optional) Arguments that ``get_miner`` takes.
    """
    miner = ItemMiner(**kwargs)
    try:
        miner.loop.add_signal_handler(signal.SIGINT, miner.close)
        miner.loop.run_until_complete(miner.mine_items(identifiers, params, callback))
    except RuntimeError:
        pass


def configure(username=None, password=None, overwrite=None):
    """Configure IA Mine with your Archive.org credentials."""
    username = input('Email address: ') if not username else username
    password = getpass('Password: ') if not password else password
    write_config_file(username, password, overwrite)



import random


def make_url(path, protocol=None, hosts=None):
    """Make an URL given a path, and optionally, a protocol and set of
    hosts to select from randomly.

    :param path: The Archive.org path.
    :type path: str

    :param protocol: (optional) The HTTP protocol to use. "https://" is
                     used by default.
    :type protocol: str

    :param hosts: (optional) A set of hosts. A host will be chosen at
                  random. The default host is "archive.org".
    :type hosts: iterable

    :rtype: str
    :returns: An Absolute URI.
    """
    protocol = 'https://' if not protocol else protocol
    host = hosts[random.randrange(len(hosts))] if hosts else 'archive.org'
    return protocol + host + path.strip()


def metadata_urls(identifiers, protocol=None, hosts=None):
    """An Archive.org metadata URL generator.

    :param identifiers: A set of Archive.org identifiers for which to
                        make metadata URLs.
    :type identifiers: iterable

    :param protocol: (optional) The HTTP protocol to use. "https://" is
                     used by default.
    :type protocol: str

    :param hosts: (optional) A set of hosts. A host will be chosen at
                  random. The default host is "archive.org".
    :type hosts: iterable

    :returns: A generator yielding Archive.org metadata URLs.
    """
    for identifier in identifiers:
        path = '/metadata/{}'.format(identifier)
        url = make_url(path, protocol, hosts)
        yield url



#!/usr/bin/env python3
"""Concurrently retrieve metadata from Archive.org items.

usage: ia-mine (<itemlist> | -) [--debug] [--workers WORKERS] [--cache]
               [--retries RETRIES] [--secure] [--hosts HOSTS]
       ia-mine [--all | --search QUERY] [[--info | --info --field FIELD...]
               |--num-found | --mine-ids | --field FIELD... | --itemlist]
               [--debug] [--rows ROWS] [--workers WORKERS] [--cache]
               [--retries RETRIES] [--secure] [--hosts HOSTS]
       ia-mine [-h | --version | --configure]

positional arguments:
  itemlist              A file containing Archive.org identifiers, one per
                        line, for which to retrieve metadata from. If no
                        itemlist is provided, identifiers will be read from
                        stdin.

optional arguments:
  -h, --help            Show this help message and exit.
  -v, --version         Show program's version number and exit.
  --configure           Configure ia-mine to use your Archive.org credentials.
  -d, --debug           Turn on verbose logging [default: False]
  -a, --all             Mine all indexed items.
  -s, --search QUERY    Mine search results. For help formatting your query,
                        see: https://archive.org/advancedsearch.php
  -m, --mine-ids        Mine items returned from search results.
                        [default: False]
  -i, --info            Print search result response header to stdout and exit.
  -f, --field FIELD     Fields to include in search results.
  -i, --itemlist        Print identifiers only to stdout. [default: False]
  -n, --num-found       Print the number of items found for the given search
                        query.
  --rows ROWS           The number of rows to return for each request made to
                        the Archive.org Advancedsearch API. On slower networks,
                        it may be useful to use a lower value, and on faster
                        networks, a higher value. [default: 50]
  -w, --workers WORKERS
                        The maximum number of tasks to run at once.
                        [default: 100]
  -c, --cache           Cache item metadata on Archive.org. Items are not
                        cached are not cached by default.
  -r, --retries RETRIES
                        The maximum number of retries for each item.
                        [default: 10]
  --secure              Use HTTPS. HTTP is used by default.
  -H, --hosts HOSTS     A file containing a list of hosts to shuffle through.

"""
from .utils import suppress_interrupt_messages, suppress_brokenpipe_messages, handle_cli_exceptions
suppress_interrupt_messages()
suppress_brokenpipe_messages()
handle_cli_exceptions()

import logging
import os
import sys
import json

from docopt import docopt, DocoptExit
from schema import Schema, Use, Or, SchemaError

from .api import mine_items, search, configure
from . import __version__
from .exceptions import AuthenticationError


asyncio_logger = logging.getLogger('asyncio')
asyncio_logger.setLevel(logging.CRITICAL)


def print_itemlist(resp):
    j = yield from resp.json(encoding='utf-8')
    for doc in j.get('response', {}).get('docs', []):
        print(doc.get('identifier'))


def main(argv=None, session=None):
    # If ia-wrapper calls main with argv argument, strip the
    # "mine" subcommand from args.
    argv = argv[1:] if argv else sys.argv[1:]

    # Catch DocoptExit error and write to stderr manually.
    # Otherwise error's vanish if executed from a pex binary.
    try:
        args = docopt(__doc__, version=__version__, argv=argv, help=True)
    except DocoptExit as exc:
        sys.exit(sys.stderr.write('{}\n'.format(exc.code)))

    # Validate args.
    open_file_or_stdin = lambda f: sys.stdin if (f == '-') or (not f) else open(f)
    parse_hosts = lambda f: [x.strip() for x in open(f) if x.strip()]
    schema = Schema({object: bool,
        '--search': Or(None, Use(str)),
        '--field': list,
        '--rows': Use(int,
            error='"{}" should be an integer'.format(args['--rows'])),
        '--hosts': Or(None, Use(parse_hosts,
            error='"{}" should be a readable file.'.format(args['--hosts']))),
        '--retries': Use(int, '"{}" should be an integer.'.format(args['--retries'])),
        '<itemlist>': Use(open_file_or_stdin,
            error='"{}" should be readable'.format(args['<itemlist>'])),
        '--workers': Use(int,
            error='"{}" should be an integer.'.format(args['--workers'])),
    })
    try:
        args = schema.validate(args)
    except SchemaError as exc:
        sys.exit(sys.stderr.write('{0}error: {1}\n'.format(__doc__, str(exc))))

    # Configure.
    if args['--configure']:
        sys.stdout.write(
            'Enter your Archive.org credentials below to configure ia-mine.\n\n')
        try:
            configure(overwrite=True)
        except AuthenticationError as exc:
            sys.stdout.write('\n')
            sys.stderr.write('error: {}\n'.format(str(exc)))
            sys.exit(1)
        sys.exit(0)

    # Search.
    if args['--search'] or args['--all']:
        query = 'all:1' if not args['--search'] else args['--search']
        callback = print_itemlist if args['--itemlist'] else None
        info_only = True if args['--info'] or args['--num-found'] else False
        params = {
            'rows': args['--rows']
        }
        for i, f in enumerate(args['--field']):
            params['fl[{}]'.format(i)] = f
        r = search(query,
                params=params,
                callback=callback,
                mine_ids=args['--mine-ids'],
                info_only=info_only,
                max_tasks=args['--workers'],
                retries=args['--retries'],
                secure=args['--secure'],
                hosts=args['--hosts'],
                debug=args['--debug'])
        if args['--info']:
            sys.stdout.write('{}\n'.format(json.dumps(r)))
        elif args['--num-found']:
            sys.stdout.write('{}\n'.format(r.get('numFound', 0)))
        sys.exit(0)

    # Mine.
    else:
        # Exit with 2 if stdin appears to be empty.
        if args['-']:
            if (not os.fstat(sys.stdin.fileno()).st_size > 0) and (sys.stdin.seekable()):
                sys.exit(2)

        mine_items(args['<itemlist>'],
                   max_tasks=args['--workers'],
                   retries=args['--retries'],
                   secure=args['--secure'],
                   hosts=args['--hosts'],
                   debug=args['--debug'])


if __name__ == '__main__':
    main()



import sys
import signal

from .exceptions import AuthenticationError


def suppress_interrupt_messages():
    """Register a new excepthook to suppress KeyboardInterrupt
    exception messages, and exit with status code 130.

    """
    old_excepthook = sys.excepthook

    def new_hook(type, value, traceback):
        if type == KeyboardInterrupt:
            sys.exit(130)
        else:
            old_excepthook(type, value, traceback)

    sys.excepthook = new_hook


def suppress_brokenpipe_messages():
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)


def handle_cli_exceptions():
    old_excepthook = sys.excepthook

    def new_hook(type, value, traceback):
        if type == AuthenticationError:
            if str(value).startswith('The request signature we calculated'):
                print('error: Your secret key does not match your access key.',
                       'Please rerun `ia-mine --configure`.', file=sys.stderr)
            elif str(value).startswith('The AWS Access Key Id'):
                print('error: Your access key is invalid.',
                       'Please rerun `ia-mine --configure`.', file=sys.stderr)
            sys.exit(1)
        else:
            old_excepthook(type, value, traceback)

    sys.excepthook = new_hook



import os.path
import configparser
import json
import asyncio

import aiohttp

from .exceptions import AuthenticationError


@asyncio.coroutine
def _get_auth_config(username, password):
    conn = aiohttp.connector.TCPConnector(share_cookies=True)

    # Login POST data.
    data = dict(
        # Cookies will expire very quickly without remember=CHECKED.
        remember='CHECKED',
        action='login',
        username=username,
        password=password,
    )

    # Login to Archive.org and add logged-in cookies to connector.
    r = yield from aiohttp.request(
            method='POST',
            url='https://archive.org/account/login.php',
            auth=aiohttp.helpers.BasicAuth(login=username, password=password),
            data=data,
            headers={'Cookie': 'test-cookie=1'},
            connector=conn)
    r.close()

    # Archive.org returns 200 for failed authentication,
    # detect auth failure by some other means.
    if 'logged-in-user' not in conn.cookies:
        raise AuthenticationError(
            'Failed to authenticate. Please check your credentials and try again.')

    # Get S3 keys using the cookies attached to the connector
    r = yield from aiohttp.request(
            method='GET',
            url='https://archive.org/account/s3.php',
            params=dict(output_json=1),
            connector=conn)
    r.close()
    j = json.loads((yield from r.read()).decode('utf-8'))

    auth_config = {
        's3': {
            'access': j.get('key', {}).get('s3accesskey'),
            'secret': j.get('key', {}).get('s3secretkey'),
        },
        'cookies': {
            'logged-in-user': conn.cookies.get('logged-in-user').value,
            'logged-in-sig': conn.cookies.get('logged-in-sig').value,
        }
    }
    conn.close()
    return auth_config


def get_auth_config(username, password):
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(_get_auth_config(username, password))


def get_config_file():
    config = configparser.RawConfigParser()

    config_dir = os.path.expanduser('~/.config/')
    if not os.path.isdir(config_dir):
        config_file = os.path.expanduser('~/.ia')
    else:
        config_file = '{0}/ia.ini'.format(config_dir)
    config.read(config_file)

    return (config_file, config)


def write_config_file(username, password, overwrite=None):
    config_file, config = get_config_file()
    auth_config = get_auth_config(username, password)

    # S3 Keys.
    if ('s3' in config) and (not overwrite):
        config_access = config.get('s3', 'access', fallback=None)
        if not config_access:
            config['s3']['access'] = auth_config.get('s3', {}).get('access')
        config_secret = config.get('s3', 'secret', fallback=None)
        if not config_secret:
            config['s3']['secret'] = auth_config.get('s3', {}).get('secret')
    else:
        config['s3'] = auth_config.get('s3')

    # Cookies.
    cookies = auth_config.get('cookies', {})
    cookies = dict((k, v.replace('%', '%%')) for k, v in cookies.items())
    if ('cookies' in config) and (not overwrite):
        config_user = config.get('cookies', 'logged-in-user', fallback=None)
        if not config_user:
            config['cookies']['logged-in-user'] = cookies.get('logged-in-user')
        config_sig = config.get('cookies', 'logged-in-sig', fallback=None)
        if not config_sig:
            config['cookies']['logged-in-sig'] = cookies.get('logged-in-sig')
    else:
        config['cookies'] = cookies

    # Write config file.
    with open(config_file, 'w') as fh:
        os.chmod(config_file, 0o700)
        config.write(fh)


def get_config(config=None):
    _config = {} if not config else config
    config_file, config = get_config_file()
    if not os.path.isfile(config_file):
        return _config

    config_dict = {
        's3': {
            'access': config.get('s3', 'access', fallback=None),
            'secret': config.get('s3', 'secret', fallback=None),
        },
        'cookies': {
            'logged-in-user': config.get('cookies', 'logged-in-user', fallback=None),
            'logged-in-sig': config.get('cookies', 'logged-in-sig', fallback=None),
        },
    }

    return dict((k, v) for k, v in config_dict.items() if v)



import urllib.request
try:
    import ujson as json
except ImportError:
    import json
import asyncio
try:
    from asyncio import JoinableQueue as Queue
except ImportError:
    from asyncio import Queue
from copy import deepcopy
import time

import aiohttp

from .config import get_config
from .requests import MineRequest
from .urls import metadata_urls, make_url
from .exceptions import AuthenticationError


class Miner(object):

    def __init__(self,
                 loop=None,
                 max_tasks=None,
                 retries=None,
                 secure=None,
                 hosts=None,
                 params=None,
                 config=None,
                 access=None,
                 secret=None,
                 debug=None):

        # Set default values for kwargs.
        loop = asyncio.get_event_loop() if not loop else loop
        max_tasks = 100 if not max_tasks else max_tasks
        max_retries = 10 if not retries else retries
        protocol = 'http://' if not secure else 'https://'
        config = get_config(config)
        access = config.get('s3', {}).get('access', access)
        secret = config.get('s3', {}).get('secret', secret)
        debug = True if debug else False

        self.max_tasks = max_tasks
        self.max_retries = max_retries
        self.protocol = protocol
        self.hosts = hosts
        self.config = config
        self.access = access
        self.debug = debug
        self.cookies = config.get('cookies', {})

        # Asyncio/Aiohttp settings.
        self.connector = aiohttp.TCPConnector(share_cookies=True, loop=loop)
        self.connector.update_cookies(self.cookies)
        self.loop = loop
        self.q = Queue(1000, loop=self.loop)
        self.q = Queue(loop=self.loop)

        # Require valid access key!
        self.assert_s3_keys_valid(access, secret)

        # Rate limiting.
        self._max_per_second = self.get_global_rate_limit()
        self._min_interval = 1.0 / float(self._max_per_second)
        self._last_time_called = 0.0

    def close(self):
        self.connector.close()
        self.loop.stop()
        self.loop.close()

    def assert_s3_keys_valid(self, access, secret):
        url = '{}s3.us.archive.org?check_auth=1'.format(self.protocol)
        r = urllib.request.Request(url)
        r.add_header('Authorization', 'LOW {0}:{1}'.format(access, secret))
        f = urllib.request.urlopen(r)
        j = json.loads(f.read().decode('utf-8'))
        if j.get('authorized') is not True:
            raise AuthenticationError(j.get('error'))

    def get_global_rate_limit(self):
        """Get the global rate limit per client.

        :rtype: int
        :returns: The global rate limit for each client.
        """
        r = urllib.request.urlopen('https://archive.org/metadata/iamine-rate-limiter')
        j = json.loads(r.read().decode('utf-8'))
        return int(j.get('metadata', {}).get('rate_per_second', 300))

    def _rate_limited():
        """A rate limit decorator for limiting the number of times the
        decorated :class:`Miner` method can be called. Limits are set in
        :attr:`Miner._max_per_second`.
        """
        def decorate(func):
            def rate_limited_func(self, *args, **kwargs):
                elapsed = time.monotonic() - self._last_time_called
                self.left_to_wait = self._min_interval - elapsed
                if self.left_to_wait > 0:
                    time.sleep(self.left_to_wait)
                func(self, *args, **kwargs)
                self._last_time_called = time.monotonic()
                yield from func(self, *args, **kwargs)
            return rate_limited_func
        return decorate

    @_rate_limited()
    def make_rate_limited_request(self, request):
        yield from request.make_request()

    @asyncio.coroutine
    def work(self):
        while True:
            request = yield from self.q.get()
            yield from self.make_rate_limited_request(request)
            self.q.task_done()

    @asyncio.coroutine
    def q_requests(self, requests):
        for req in requests:
            self.q.put_nowait(req)

    @asyncio.coroutine
    def mine(self, requests):
        workers = [asyncio.Task(self.work(), loop=self.loop)
                   for _ in range(self.max_tasks)]
        asyncio.Task(self.q_requests(requests))

        yield from self.q.join()
        yield from asyncio.sleep(1)
        while not self.q.empty():
            yield from asyncio.sleep(1)

        for w in workers:
            w.cancel()


class ItemMiner(Miner):

    def __init__(self, **kwargs):
        super(ItemMiner, self).__init__(**kwargs)

    @asyncio.coroutine
    def mine_items(self, identifiers, params=None, callback=None):
        """Mine metadata from Archive.org items.

        :param identifiers: Archive.org identifiers to be mined.
        :type identifiers: iterable

        :param params: URL parameters to send with each metadata
                       request.
        :type params: dict

        :param callback: A callback function to be called on each
                         :py:class:`aiohttp.client.ClientResponse`.
        :type callback: func
        """
        # By default, don't cache item metadata in redis.
        params = {'dontcache': 1} if not params else {}
        requests = metadata_requests(identifiers, params, callback, self)
        yield from self.mine(requests)


class SearchMiner(ItemMiner):

    def __init__(self, **kwargs):
        super(SearchMiner, self).__init__(**kwargs)
        # Item mining queue.
        self.iq = Queue(1000, loop=self.loop)

    def get_search_params(self, query, params):
        default_rows = 500
        search_params = {
            'q': 'all:1',
            'page': 1,
            'output': 'json',
        }
        if params:
            search_params.update({k: v for k, v in params.items() if v})
        if query:
            search_params['q'] = query
        if 'rows' not in search_params:
            search_params['rows'] = default_rows
        return search_params

    def get_search_info(self, params):
        url = make_url('/advancedsearch.php?', self.protocol, self.hosts)
        p = deepcopy(params)
        p['rows'] = 0

        params = urllib.parse.urlencode(p)
        url += params
        f = urllib.request.urlopen(url)
        return json.loads(f.read().decode('utf-8'))

    @asyncio.coroutine
    def _handle_search_results(self, resp, params=None, callback=None):
        j = yield from resp.json(encoding='utf-8')
        resp.close()
        identifiers = []
        for doc in j.get('response', {}).get('docs', []):
            if not doc.get('identifier'):
                continue
            identifiers.append(doc['identifier'])
        for req in metadata_requests(identifiers, params, callback, self):
            self.iq.put_nowait(req)

    def search_requests(self, query=None, params=None, callback=None, mine_ids=None):
        """Mine Archive.org search results.

        :param query: The Archive.org search query to yield results for.
                      Refer to https://archive.org/advancedsearch.php#raw
                      for help formatting your query.
        :type query: str

        :param params: The URL parameters to send with each request sent
                       to the Archive.org Advancedsearch Api.
        :type params: dict
        """
        # If mining ids, devote half the workers to search and half to item mining.
        if mine_ids:
            self.max_tasks = self.max_tasks/2
        # When mining id's, the only field we need returned is "identifier".
        if mine_ids and params:
            params = dict((k, v) for k, v in params.items() if 'fl' not in k)
            params['fl[]'] = 'identifier'

        # Make sure "identifier" is always returned in search results.
        fields = [k for k in params if 'fl' in k]
        if (len(fields) == 1) and (not any('identifier' == params[k] for k in params)):
            # Make sure to not overwrite the existing fl[] key.
            i = 0
            while params.get('fl[{}]'.format(i)):
                i += 1
            params['fl[{}]'.format(i)] = 'identifier'

        search_params = self.get_search_params(query, params)
        url = make_url('/advancedsearch.php', self.protocol, self.hosts)

        search_info = self.get_search_info(search_params)
        total_results = search_info.get('response', {}).get('numFound', 0)
        total_pages = (int(total_results/search_params['rows']) + 1)

        for page in range(1, (total_pages + 1)):
            params = deepcopy(search_params)
            params['page'] = page
            if not callback and mine_ids:
                callback = self._handle_search_results
            req = MineRequest('GET', url, self.access,
                              callback=callback,
                              max_retries=self.max_retries,
                              debug=self.debug,
                              params=params,
                              connector=self.connector)
            yield req

    @asyncio.coroutine
    def mine_items(self):
        while True:
            request = yield from self.iq.get()
            yield from self.make_rate_limited_request(request)
            self.iq.task_done()

    @asyncio.coroutine
    def search(self, query=None, params=None, callback=None, mine_ids=None):
        search_requests = self.search_requests(query, params, callback, mine_ids)
        if mine_ids:
            workers = [asyncio.Task(self.mine_items(), loop=self.loop)
                       for _ in range(self.max_tasks)]

        yield from self.mine(search_requests)
        # Wait a bit for all connections to close.
        yield from asyncio.sleep(1)

        if mine_ids:
            for w in workers:
                w.cancel()


# metadata_requests() ____________________________________________________________________
def metadata_requests(identifiers, params=None, callback=None, miner=None):
    protocol = None if not miner else miner.protocol
    hosts = None if not miner else miner.hosts
    urls = metadata_urls(identifiers, protocol, hosts)

    for url in urls:
        yield MineRequest('GET', url, miner.access,
                          callback=callback,
                          max_retries=miner.max_retries,
                          debug=miner.debug,
                          params=params,
                          connector=miner.connector)
